#include <linux/init.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/module.h>
#include <asm/uaccess.h>


#ifndef MAJORNUM
#define	MAJORNUM	250
#endif

static	int	major = MAJORNUM;
static	char	device_name [] = "fake";


static int fake_open (struct inode *inode, struct file *file)
{
	int	i;
	struct input_dev *dev;

	if (!(dev = kmalloc (2 * sizeof *dev, GFP_KERNEL)))
		return -ENOMEM;
	memset (dev, 0, 2 * sizeof *dev);

	dev [0].name = "Bluetooth keyboard";
	dev [0].evbit [0] = BIT(EV_KEY) | BIT(EV_REP);
	for (i = KEY_RESERVED; i < BTN_MISC; i++)
		set_bit (i, dev [0].keybit);

	dev [1].name = "Bluetooth mouse";
	dev [1].evbit [0] = BIT(EV_KEY) | BIT(EV_REL);
	for (i = BTN_LEFT; i <= BTN_BACK; i++)
		set_bit (i, dev [1].keybit);
	set_bit (REL_X, dev [1].relbit);
	set_bit (REL_Y, dev [1].relbit);
	set_bit (REL_WHEEL, dev [1].relbit);

	input_register_device (dev);
	input_register_device (dev+1);
	file->private_data = dev;
	return 0;
}

static int fake_release (struct inode *inode, struct file *file)
{
	struct input_dev *dev = (struct input_dev *) file->private_data;

	if (dev) {
		file->private_data = NULL;
		input_unregister_device (dev);
		input_unregister_device (dev+1);
		kfree (dev);
	}
	return 0;
}

static ssize_t fake_write (struct file *file, const char *buffer,
			   size_t count, loff_t *ppos)
{
#define	BUFSZ (4*64)
	int	i, wrote = 0, amt, evt, which, val;
	char	buf [BUFSZ];
	struct input_dev *dev = (struct input_dev *) file->private_data;

	for (wrote = 0; wrote < count; wrote += amt) {
		amt = count - wrote;
		if (amt > BUFSZ)
			amt = BUFSZ;
		if (copy_from_user (buf, buffer + wrote, amt))
			return -EFAULT;
		for (i = 0; i < amt - 3; i += 4) {
			evt = buf [i];
			which = (buf [i+2] & 0xff) << 8 |
				buf [i+1] & 0xff;
			val = buf [i+3];
			input_event (dev + (evt == EV_REL ||
					    evt == EV_KEY &&
					    which >= BTN_MISC),
				     evt, which, val);
		}
	}

	return wrote;
}


static devfs_handle_t devfs_handle;

static struct file_operations fops = {
	owner: THIS_MODULE,
	open: fake_open,
	release: fake_release,
	write: fake_write
};

static int __init driver_init (void)
{
	if (devfs_register_chrdev (major, device_name, &fops)) {
		printk (KERN_ERR "%s: unable to register char major %d",
			device_name, major);
		return -EBUSY;
	}
	devfs_handle = devfs_mk_dir (NULL, device_name, NULL);
	return 0;
}

static void __exit driver_exit (void)
{
	devfs_unregister (devfs_handle);
        if (devfs_unregister_chrdev (major, device_name))
                printk (KERN_ERR "%s: can't unregister char major %d",
			device_name, major);
}

module_init(driver_init);
module_exit(driver_exit);

MODULE_AUTHOR("Peter Klausler <pmk@cray.com>");
MODULE_DESCRIPTION("Fake input device");
/*MODULE_PARM(major,"i");*/
MODULE_LICENSE("GPL");
